#include "block.h"
#include <QJsonArray>
#include <QTimer>
#include <layout.h>
#include "nfcuid.h"


Block::Block(OverlordItemStore* items, uint32_t id):
	items_(items), id_(id)
{

}

void Block::addBorder(std::shared_ptr<BlockBorder> border)
{
	borders_.push_back(border);
	std::shared_ptr<Block> block = border->getOtherBlock(this);
	if(!block || block.get() == this)
		return;
	connect(block.get(), &Block::blockedChanged, this, &Block::checkWaits);
}

bool Block::blocked()
{
	if(tags_.empty() && waits_.empty())
		return false;
	return true;
}

bool Block::ownsTag(const NfcUid& tag)
{
	for(const NfcUid& candiate : tags_)
	{
		if(candiate == tag)
			return true;
	}
	return false;
}

bool Block::tagIsOfOwnedTrain(const NfcUid& tag)
{
	std::shared_ptr<Train> train = items_->getTrainThatOwns(tag);
	if(!train)
		return false;
	for(const NfcUid& canidate : tags_)
	{
		std::shared_ptr<Train> canidateTrain = items_->getTrainThatOwns(canidate);
		if(*canidateTrain == *train)
			return true;
	}
	return false;
}

bool Block::ownsBorder(std::shared_ptr<BlockBorder> border)
{
	if(!border)
		return false;
	for(std::shared_ptr<BlockBorder> borderTest : borders_)
	{
		if(borderTest == border)
			return true;
	}
	return false;
}

void Block::checkWaits(bool isBlocked)
{
	if(isBlocked)
		return;
	bool wasBlocked = blocked();
	for(std::vector<Wait>::iterator iter = waits_.begin(); iter != waits_.end();)
	{
		if(iter->type != WAIT_TYPE_BLOCK)
		{
			++iter;
			continue;
		}

		std::shared_ptr<Block> block = iter->targetBlock.lock();
		std::shared_ptr<Train> train = iter->train.lock();
		qDebug()<<__func__<<"trying to push train"<<train->getTrainId()<<"to block"<<block->id();
		if(block && train && train->suspendend() && block->pushTag(iter->tag))
		{
			train->resume();
			std::shared_ptr<BlockBorder> border = iter->border.lock();
			if(border)
				border->updateSignals();
			iter = waits_.erase(iter);
		}
		else if(!block || !train || !train->suspendend())
		{
			iter = waits_.erase(iter);
		}
		else
		{
			++iter;
		}
	}
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

bool Block::pushTag(const NfcUid& tag)
{
	if(!blocked() || tagIsOfOwnedTrain(tag))
	{
		addTag(tag);
		return true;
	}
	return false;
}

void Block::removeTraverse(std::shared_ptr<Train> train)
{
	bool wasBlocked = blocked();
	for(std::vector<Wait>::iterator iter = waits_.begin(); iter != waits_.end();)
	{
		if(iter->type == WAIT_TYPE_TIMED_TRAVERSE)
		{
			std::shared_ptr<Train> iterTrain = iter->train.lock();
			if(!iterTrain)
			{
				iter = waits_.erase(iter);
				continue;
			}
			if(*iterTrain == *train)
			{
				qDebug()<<"Train"<<iterTrain->getTrainId()<<"removed from traverse wait for block"<<id_;
				iter = waits_.erase(iter);
				continue;
			}
		}
		 ++iter;
	}
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

void Block::addTraverseWait(const NfcUid& tag)
{
	std::shared_ptr<Train> train = items_->getTrainThatOwns(tag);
	if(!train)
		return;
	if(!train->hasBackTag())
	{
		qDebug()<<"Train"<<train->getTrainId()<<"added as timed traverse wait for block"<<id_;
		Wait wait;
		wait.tag = tag;
		wait.train = train;
		wait.type = WAIT_TYPE_TIMED_TRAVERSE;
		waits_.push_back(wait);
		QTimer::singleShot(5000, this, [this, train](){removeTraverse(train);});
	}
}

void Block::addBlockWait(const NfcUid& tag, std::shared_ptr<Block> block, std::shared_ptr<BlockBorder> border)
{
	std::shared_ptr<Train> train = items_->getTrainThatOwns(tag);
	if(!train)
		return;
	Wait wait;
	wait.type = WAIT_TYPE_BLOCK;
	wait.train = train;
	wait.direction = train->getDirection();
	wait.targetBlock = block;
	wait.border = border;
	wait.tag = tag;
	connect(train.get(), &Train::unsuspended, this, &Block::unsuspendedTrain);
	waits_.push_back(wait);
	qDebug()<<"Train"<<train->getTrainId()<<"is wating at border for block"<<block->id();
}

void Block::addTag(const NfcUid& tag)
{
	bool wasBlocked = blocked();
	tags_.push_back(tag);
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

void Block::unsuspendedTrain(uint32_t id, int direction)
{
	qDebug()<<"Train with item id"<<id<<"reports unsuspended";
	bool wasBlocked = blocked();
	for(std::vector<Wait>::iterator iter = waits_.begin(); iter != waits_.end(); ++iter)
	{
		std::shared_ptr<Train> workTrain = iter->train.lock();
		if(!workTrain)
		{
			waits_.erase(iter);
			unsuspendedTrain(id, direction);
			break;
		}
		if(workTrain->id() == id)
		{
			std::shared_ptr<Block> block = iter->targetBlock.lock();
			disconnect(workTrain.get(), &Train::unsuspended, this, &Block::unsuspendedTrain);
			if(block)
				block->addTag(iter->tag);
			waits_.erase(iter);
			break;
		}
	}
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

void Block::removeTag(const NfcUid& tag)
{
	for(std::vector<NfcUid>::iterator iter = tags_.begin(); iter != tags_.end(); ++iter)
	{
		if(*iter == tag)
		{
			tags_.erase(iter);
			break;
		}
	}
}

void Block::purgeTag(const NfcUid& tag)
{
	bool wasBlocked = blocked();
	removeTag(tag);
	std::shared_ptr<Train> train = items_->getTrainThatOwns(tag);
	if(train)
		removeTraverse(train);
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

void Block::tagArrivedAtBorder(NfcUid tag, std::weak_ptr<BlockBorder> borderIn)
{
	bool wasBlocked = blocked();
	qDebug()<<__func__<<"block"<<id();
	std::shared_ptr<BlockBorder> border = borderIn.lock();
	if(!border)
	{
		qWarning()<<"border has expired in"<<__func__;
		return;
	}

	if(!ownsBorder(border))
		return;

	qDebug()<<"Block"<<id_<<"owns subject border";

	if(!ownsTag(tag))
		return;

	qDebug()<<"Block"<<id_<<"owns subject tag";

	std::shared_ptr block = border->getOtherBlock(this);
	if(block)
	{
		if(!block->pushTag(tag))
		{
			std::shared_ptr<Train> train = items_->getTrainThatOwns(tag);
			train->suspend();
			addBlockWait(tag, block, border);
		}
		else
		{
			addTraverseWait(tag);
		}
		removeTag(tag);
	}
	if(wasBlocked != blocked())
		blockedChanged(!wasBlocked);
}

void Block::store(QJsonObject &json)
{
	json["Id"] = static_cast<double>(id_);
	QJsonArray borders;
	for(auto& border : borders_)
	{
		QJsonObject borderObject;
		borderObject["BorderId"] =  static_cast<double>(border->id());
		borders.append(borderObject);
	}
	json["Borders"] = borders;
}

void Block::load(const QJsonObject &json)
{
	id_ = static_cast<uint32_t>(json["Id"].toDouble(0));
}

void Block::populate(const QJsonObject& json, Layout* layout)
{
	borders_.clear();
	const QJsonArray bordersArray(json["Borders"].toArray(QJsonArray()));
	for(const auto& arrayObj : bordersArray)
	{
		if(arrayObj.isObject())
		{
			uint32_t borderId = static_cast<uint32_t>(arrayObj.toObject()["BorderId"].toDouble(INT32_MAX));
			std::shared_ptr<BlockBorder> border = layout->getBorder(borderId);
			if(border)
				addBorder(border);
			else if(borderId == INT32_MAX)
				qWarning()<<"BorderId field is missing in border array for block"<<id_;
			else
				qWarning()<<"border"<<borderId<<"dosent exist but is needed by block"<<id_;
		}
	}
}

