#include "blockborder.h"
#include <QJsonArray>
#include <layout.h>

BlockBorder::BlockBorder(uint8_t reader, std::pair<std::weak_ptr<Block>, std::weak_ptr<Block>> blocks, uint32_t id): reader_(reader), id_(id)
{
	setBlocks(blocks);
}

bool BlockBorder::isReader(uint8_t reader)
{
	return reader == reader_;
}

uint8_t BlockBorder::getReader()
{
	return reader_;
}

void BlockBorder::setBlocks(std::pair<std::weak_ptr<Block>, std::weak_ptr<Block>> blocks)
{
	blocks_ = blocks;
	std::shared_ptr<Block> first = blocks.first.lock();
	if(first)
		connect(first.get(), &Block::blockedChanged, this, &BlockBorder::blockedChanged);
	std::shared_ptr<Block> second = blocks.second.lock();
	if(second)
		connect(second.get(), &Block::blockedChanged, this, &BlockBorder::blockedChanged);
}

void BlockBorder::setSignal(std::weak_ptr<Signal> signalWeak, int8_t value)
{
	std::shared_ptr<Signal> signal = signalWeak.lock();
	if(signal)
	{
		signal->setValue(value);
	}
}

void BlockBorder::blockedChanged(bool blocked)
{
	(void)blocked;
	updateSignals();
}

void BlockBorder::updateSignals()
{
	for(size_t i = 0; i < signals_.size(); ++i)
	{
		std::shared_ptr<Signal> signal = signals_[i].lock();
		if(signal)
		{
			std::shared_ptr<Block> block = signalDirections_[i] ? blocks_.first.lock() : blocks_.second.lock();
			if(block && block->blocked())
			{
				if(signal->getValue() != Signal::STOP)
				{
					std::weak_ptr<Signal> signalWeak = signal;
					QTimer::singleShot(500, signal.get(), [signalWeak](){BlockBorder::setSignal(signalWeak, Signal::STOP);});
				}
			}
			else
			{
				if(signal->getValue() != Signal::GO)
				{
					std::weak_ptr<Signal> signalWeak = signal;
					QTimer::singleShot(500, signal.get(), [signalWeak](){BlockBorder::setSignal(signalWeak, Signal::GO);});
				}
			}
		}
		else
		{
			signals_.erase(signals_.begin()+i);
			updateSignals();
			break;
		}
	}
}

void BlockBorder::addSignal(std::weak_ptr<Signal> signal, std::shared_ptr<Block> block)
{
	signals_.push_back(signal);
	signalDirections_.push_back(block == blocks_.first.lock());
}

std::shared_ptr<Block> BlockBorder::getOtherBlock(Block* block)
{
	std::shared_ptr<Block> first = blocks_.first.lock();
	std::shared_ptr<Block> second = blocks_.second.lock();

	if(first && first.get() == block)
		return second;
	if(second && second.get() == block)
		return first;
	return std::shared_ptr<Block>();
}

std::shared_ptr<Block> BlockBorder::getOverlap(BlockBorder* border)
{
	if(border->getBlocks().first.lock() == blocks_.first.lock() || border->getBlocks().first.lock() == blocks_.second.lock())
	{
		return border->getBlocks().first.lock();
	}
	else if(border->getBlocks().second.lock() == blocks_.first.lock() || border->getBlocks().second.lock() == blocks_.second.lock())
	{
		return border->getBlocks().second.lock();
	}
	return std::shared_ptr<Block>();
}

void BlockBorder::store(QJsonObject& json)
{
	json["Id"] = static_cast<double>(id_);
	json["Reader"] = static_cast<double>(reader_);
	QJsonArray signalArray;
	for(size_t i = 0; i < signals_.size(); ++i)
	{
		std::shared_ptr<Signal> signalPtr = signals_[i].lock();
		if(signalPtr)
		{
			QJsonObject signalObject;
			signalObject["SignalId"] = static_cast<double>(signalPtr->getSignalId());
			signalObject["Direction"] = static_cast<double>(signalDirections_[i]);
			signalArray.append(signalObject);
		}
	}
	json["Signals"] = signalArray;
}

void BlockBorder::load(const QJsonObject& json)
{
	id_ = static_cast<uint32_t>(json["Id"].toDouble(0));
	reader_ = json["Reader"].toDouble(0);
	const QJsonArray signalArray(json["Signals"].toArray(QJsonArray()));
	for(const auto& arrayObj : signalArray)
	{
		if(arrayObj.isObject())
		{
			QJsonObject arrayObjObj = arrayObj.toObject();
			signalIds_.push_back(std::pair<uint32_t, bool>(static_cast<uint32_t>(arrayObjObj["SignalId"].toDouble(INT32_MAX)), static_cast<bool>(arrayObjObj["Direction"].toDouble(0))));
		}
	}
}

bool BlockBorder::hasMissingSingal()
{
	for(auto& signal : signalIds_)
	{
		qDebug()<<"Missing signal id:"<<signal.first<<"Direction:"<<signal.second;
	}
	return !signalIds_.empty();
}

void BlockBorder::informOfItem(std::weak_ptr<Item> item)
{
	std::shared_ptr<Item> workItem = item.lock();
	if(!workItem)
		return;
	std::shared_ptr<Signal> signal = std::dynamic_pointer_cast<Signal>(workItem);
	if(!signal)
		return;
	for(const std::pair<uint32_t, bool>& signalPair : signalIds_)
	{
		if(signalPair.first == signal->getSignalId())
		{
			signals_.push_back(signal);
			signalDirections_.push_back(signalPair.second);
			qDebug()<<"Border"<<id_<<"got Signal"<<signal->getSignalId();
		}
	}
	std::erase_if(signalIds_, [signal](std::pair<uint32_t, bool> pair) -> bool{return pair.first == signal->getSignalId();});
}
