#include "layout.h"
#include <memory>
#include <QJsonArray>

Layout::Layout(OverlordItemStore* items, QObject *parent): QObject{parent}, items_(items)
{
}

void Layout::removeTagFromAllBlocks(const NfcUid& tag)
{
	for(std::shared_ptr<Block> block : blocks_)
	{
		block->purgeTag(tag);
	}
}

void Layout::registerTagInLimbo(const NfcUid& tag, std::shared_ptr<BlockBorder> border)
{
	for(auto iter = limbo_.begin(); iter != limbo_.end(); ++iter)
	{
		NfcUid limboTag = iter->first;
		std::shared_ptr<BlockBorder> limboBorder = iter->second.lock();
		if(!limboBorder)
		{
			limbo_.erase(iter);
			registerTagInLimbo(tag, border);
			return;
		}

		if(limboTag == tag)
		{
			if(border == limboBorder)
				return;

			std::shared_ptr<Block> overlap = border->getOverlap(limboBorder.get());
			if(overlap)
			{
				qDebug()<<"Tag"<<tag.toString().c_str()<<"removed from limbo and added to block"<<overlap->id()<<"while crossing border"<<border->id();
				overlap->addTag(tag);
				overlap->tagArrivedAtBorder(tag, border);
				limbo_.erase(iter);
				return;
			}
			else
			{
				iter->second = border;
				return;
			}

		}
	}
	std::shared_ptr<Block> first = border->getBlocks().first.lock();
	std::shared_ptr<Block> second = border->getBlocks().second.lock();

	qDebug()<<"Tag"<<tag.toString().c_str()<<"added to limbo between block"<<
			  (first ? QString::number(first->id()) : "invalid")<<"and block"<<(second ? QString::number(second->id()) : "invalid");
	limbo_.push_back(std::pair<NfcUid, std::weak_ptr<BlockBorder>>(tag, border));
}

void Layout::tagArrivedAtReader(uint8_t reader, NfcUid tag)
{
	std::shared_ptr<BlockBorder> border;
	for(std::shared_ptr<BlockBorder> borderTest : borders_)
	{
		if(borderTest->getReader() == reader)
		{
			border = borderTest;
			break;
		}
	}
	if(!border)
	{
		qWarning()<<"reader "<<reader<<"is not registerd with any border";
		return;
	}

	std::shared_ptr<Train> train = items_->getTrainThatOwns(tag);
	qDebug()<<"Train"<<(train ? train->getTrainId() : -1)<<"Tag"<<tag.toString().c_str()<<"arrived at border"<<border->id();

	bool tagHandled = false;
	for(std::shared_ptr<Block>& block : blocks_)
	{
		if(block->ownsTag(tag))
		{
			tagHandled = true;
			if(block->ownsBorder(border))
			{
				block->tagArrivedAtBorder(tag, border);
			}
			else
			{
				removeTagFromAllBlocks(tag);
				registerTagInLimbo(tag, border);
			}
			break;
		}
	}

	if(!tagHandled)
		registerTagInLimbo(tag, border);
}

void Layout::addBlock(std::shared_ptr<Block> block)
{
	blocks_.push_back(block);
}

void Layout::addBorder(std::shared_ptr<BlockBorder> border)
{
	borders_.push_back(border);
}

std::shared_ptr<BlockBorder> Layout::getBorder(uint32_t id)
{
	for(const std::shared_ptr<BlockBorder>& border : borders_)
	{
		if(border->id() == id)
			return border;
	}
	return std::shared_ptr<BlockBorder>();
}

std::shared_ptr<Block> Layout::getBlock(uint32_t id)
{
	for(const std::shared_ptr<Block>& block : blocks_)
	{
		if(block->id() == id)
			return block;
	}
	return std::shared_ptr<Block>();
}

void Layout::store(QJsonObject& json)
{
	QJsonArray blockArray;
	for(const std::shared_ptr<Block>& block : blocks_)
	{
		QJsonObject object;
		block->store(object);
		blockArray.append(object);
	}
	json["Blocks"] = blockArray;

	QJsonArray borderArray;
	for(const std::shared_ptr<BlockBorder>& border : borders_)
	{
		QJsonObject object;
		border->store(object);
		borderArray.append(object);
	}
	json["Borders"] = borderArray;
}

void Layout::load(const QJsonObject& json)
{
	std::vector<QJsonObject> blockJsonObjects;
	const QJsonArray blockArray(json["Blocks"].toArray(QJsonArray()));
	for(const auto& object : blockArray)
	{
		if(object.isObject())
		{
			const QJsonObject jsonObject = object.toObject();
			blockJsonObjects.push_back(jsonObject);
			std::shared_ptr<Block> block(new Block(items_));
			block->load(jsonObject);
			blocks_.push_back(block);
		}
	}

	std::vector<QJsonObject> borderJsonObjects;
	const QJsonArray borderArray(json["Borders"].toArray(QJsonArray()));
	for(const auto& object : borderArray)
	{
		if(object.isObject())
		{
			const QJsonObject jsonObject = object.toObject();
			borderJsonObjects.push_back(jsonObject);
			std::shared_ptr<BlockBorder> border(new BlockBorder);
			border->load(object.toObject());
			borders_.push_back(border);
		}
	}

	for(size_t i = 0; i < blocks_.size(); ++i)
	{
		blocks_[i]->populate(blockJsonObjects[i], this);
	}

	for(std::shared_ptr<BlockBorder>& border : borders_)
	{
		std::shared_ptr<Block> first;
		std::shared_ptr<Block> second;
		for(std::shared_ptr<Block>& block : blocks_)
		{
			if(block->ownsBorder(border))
			{
				if(!first)
					first = block;
				else if(!second)
					second = block;
				else
					qWarning()<<"border with id"<<border->id()<<"is assigned to more than 2 blocks";
			}
		}
		if(!first || !second)
			qWarning()<<"border with id"<<border->id()<<"is assigned to less than than 2 blocks";
		border->setBlocks(std::pair<std::weak_ptr<Block>, std::weak_ptr<Block>>(first, second));

		for(std::shared_ptr<Item>& item : *items_->getItems())
			border->informOfItem(item);

		connect(items_, &ItemStore::itemAdded, border.get(), &BlockBorder::informOfItem);
	}

	for(size_t i = 0; i < blocks_.size(); ++i)
	{
		blocks_[i]->populate(blockJsonObjects[i], this);
		if(blocks_[i]->getBorders().empty())
			qWarning()<<"block with id"<<blocks_[i]->id()<<"doset have at least one border";
	}
}
